 
/*
  
  0139-remote-433-transmitter.c
  
  
  Based on the SendDemo example from the RC Switch library
  https://github.com/sui77/rc-switch/

  https://randomnerdtutorials.com/decode-and-send-433-mhz-rf-signals-with-arduino/


  // https://raw.githubusercontent.com/damellis/attiny/ide-1.6.x-boards-manager/package_damellis_attiny_index.json

  power consumption arduino nano:
  
  28.2 mA    Normal mode (no sleep)
  16.3 mA    LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_OFF);  
             (with WDT sleep and disabled red LED)

   27.4uA    DigiSpark ATtiny85 with WDT sleep, ADC_OFF
   
    9.0uA    Attiny85      


  version 0.5
  last updated 28 December 2021

  

  HOW TO PROGRAM:
    - start arduino IDE (v1.8.9)
    - in Preferences, add this ATtiny board manager url:
    https://raw.githubusercontent.com/damellis/attiny/ide-1.6.x-boards-manager/package_damellis_attiny_index.json
    - in Tools - Board.. - Board Manager, install the 'David A Mellis' library/boards
    - select ATtiny25/45/85, ATtiny85, internal 1MHz
    - programmer: USBasp  (THIS IS NOT A USB 'COM' DEVICE, so you won't see anything under 'Ports' !!)


    FUSES: default fuses for Attiny85 @ MHz internal oscillator are usually sufficient: L=62  H=DF   E=FF
           However, if Brown-Out-Detection is required, set them to: L=62   H=DD   E=FF
           NOTE: also disable the code that manually overrides and disables the BOD near line 245.


   Changelog:

   v0.5
     - replaced Arduino DallasTemperature and OneWire libraries with my own.

         With my own DS18B20 code
           Sketch uses 3716 bytes (45%) of program storage space. Maximum is 8192 bytes.
           Global variables use 204 bytes (39%) of dynamic memory, leaving 308 bytes for local variables. Maximum is 512 bytes.

     - removed EEPROM device-storage code as it often incremented the device id for no reason
               

   v0.4
     - added DS18B20 temperature sensor support.
          add DallasTemperature library from Miles Burton (from arduino gui library manager)
     - removed car code
     - removed xor checksum to make way for temperature sensor data (alternates)

         Without DallasTemperature and OneWire:
           Sketch uses 2904 bytes (35%) of program storage space. Maximum is 8192 bytes.
           Global variables use 184 bytes (35%) of dynamic memory, leaving 328 bytes for local variables. Maximum is 512 bytes.
 
         With DallasTemperature and OneWire:
           Sketch uses 5404 bytes (65%) of program storage space. Maximum is 8192 bytes.
           Global variables use 227 bytes (44%) of dynamic memory, leaving 285 bytes for local variables. Maximum is 512 bytes.    
   
   v0.3
     - enabled BOD to make it more reliable in low-voltage conditions
     
   v0.2
     - disabled BOD to save more power


                          ATtiny85
               
                          +--- ---+
              (RST)  PB5 -|1 |_| 8|- VCC 
     FET SWITCH ---- PB3 -|2     7|- PB2    (SCK)   ---  DATA (RF OUT)
            ADC ---- PB4 -|3     6|- PB1    (MISO)  ---  LED  (apply 3.3V at boot to increment device_id)
                     GND -|4     5|- PB0    (MOSI)  ---  DS18B20
                          +-------+



  minimising footprint

  was:
    program: 3142 bytes (.text + .data + .bootloader)
    data:     188 bytes (.data + .bss + .noint)

  now:



   See https://github.com/sui77/rc-switch

   The default length of sending bits is 24 (3 Bytes, e.g. '05B9AC'), so we'll have to encode in here:
      - the device's unique serial number
      - the data payload
      - a checksum
    
   The .send command can be used like this to send these 3 bytes as a long binary string of 1's and 0's:
        
      mySwitch.send("000101010101010001010100");   // Binary code - (155454)
 

   For our code we are going to use it like so:

          --Byte1--    --Byte2--    --Byte3--
          Nib1 Nib2    Nib3 Nib4    Nib5 Nib6
          XXXX XXXX    XXXX AABB    BBBB BBBB
          ---- ----    ---- ----    ---- ----
            |    |       |   | |      |    |
            +-+--+-------+   | +--+---+----+
              |              |    |
              |              |    +----  The second half of nibble 4 and nibbles 5 + 6 sets the data value (BB)
              |              |
              |              +--  The first half of nibble 4 sets the data type (AA)
              |
              +--   Nibble 1, 2 and 3 set the device ID (XX)




*/



#define __AVR_ATtiny85__
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>



#include <RCSwitch.h>
RCSwitch mySwitch = RCSwitch();


#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define DEV_BASE 0x019    // default device address
uint16_t device_id;

uint8_t ee_addr = 1;  // device id (additions) in eeprom
#define EE_ADDR 1     // e.g. 100 + ee_addr == device_id (101 = 0x065)


// WDTIME   0=16ms, 1=32ms,2=64ms,3=128ms,4=250ms,5=500ms,6=1 sec,7=2 sec, 8=4 sec, 9= 8sec
#define WDTIME 9              // set watchdog timer sinle sleep duration
#define SLEEPCYCLES 2        // extend WDT duration by looping it several times

uint8_t sleepcycle; 
uint8_t data_type;

#define LEDPIN 1     // PB1 (pin 6) (shared with MISO)
#define TXPIN  2     // PB2 (pin 7) (shared with SCK)
#define SWITCH 3     // PB3 (pin 2)
#define ADCPIN A2    // PB4 (pin 3)


//-- begin DS18B20 defines --

#define THERM_PORT   PORTB
#define THERM_DDR    DDRB
#define THERM_PIN    PINB
#define THERM_DQ     PB0  // Define DS1820 DQ pin (pin 5) (shared with MOSI)

#define THERM_INPUT_MODE()    THERM_DDR&=~(1<<THERM_DQ)
#define THERM_OUTPUT_MODE()   THERM_DDR|=(1<<THERM_DQ)
#define THERM_LOW()           THERM_PORT&=~(1<<THERM_DQ)
#define THERM_HIGH()          THERM_PORT|=(1<<THERM_DQ)

#define HIGH   1
#define LOW    0
#define INPUT  1 
#define OUTPUT 0

#define THERM_CMD_CONVERTTEMP    0x44
#define THERM_CMD_RSCRATCHPAD    0xbe
#define THERM_CMD_WSCRATCHPAD    0x4e
#define THERM_CMD_CPYSCRATCHPAD  0x48
#define THERM_CMD_RECEEPROM      0xb8
#define THERM_CMD_RPWRSUPPLY     0xb4
#define THERM_CMD_SEARCHROM      0xf0
#define THERM_CMD_READROM        0x33
#define THERM_CMD_MATCHROM       0x55
#define THERM_CMD_SKIPROM        0xcc
#define THERM_CMD_ALARMSEARCH    0xec

#define THERM_DECIMAL_STEPS_12BIT   625    // 0.0625, for 12-bit accuracy

#define PRESENCE_ERR 0xFF
#define DATA_ERR     0xFF
#define LAST_DEVICE  0x00
#define SEARCH_FIRST 0xFF

#define CLEAR(s) memset(&(s), 0, sizeof(s))     // to clear an array

#define ROMS 1  // max number of ROMs (sensors) we can query

int      got_ack;
uint8_t  romcodes[ROMS][9];  // create arrays for all sensors
uint8_t  scratchpad[9];      // stores a scratchpad
char     buffer[16];

//--- end DS18B20 defines --





void setup() 
{
  
  // Transmitter is connected to Arduino Pin  
  mySwitch.enableTransmit(TXPIN);

  // Optional set pulse length.
  mySwitch.setPulseLength(10);
  
  // Optional set protocol (default is 1, will work for most outlets)
  mySwitch.setProtocol(1);
  
  // Optional set number of transmission repetitions.
  // mySwitch.setRepeatTransmit(15);

  pinMode(SWITCH,OUTPUT);
  digitalWrite(SWITCH,LOW);


  // increment device id if Vcc is detected on LEDPIN 
  pinMode(LEDPIN,INPUT);
  
   // use LEDPIN as output
  pinMode(LEDPIN,OUTPUT);
  digitalWrite(LEDPIN,LOW);

  device_id = DEV_BASE;
  
  data_type = 1;

}

//-- begin DS18B20 functions--

void pMode(int pin,int state)
{ 
  if (state == OUTPUT)
  {
    (DDRB |= (1 << pin)); 
  } else { 
    (DDRB &= ~(1 << pin)); 
  }
} 

void dWrite(int pin, int state)
{
  if (state == HIGH)
  { 
    (PORTB |= (1 << pin)); 
  } else {
    (PORTB &= ~(1 << pin));
  }
} 


int dRead(int pin)
{ 
  if (bit_is_set(PINB,pin))
  {
    return 1; 
  } else { 
    return 0; 
  }
} 


uint8_t owReset(int Pin)
{
  uint8_t i;

  dWrite(Pin, HIGH);
  pMode(Pin, INPUT);

  dWrite(Pin, LOW);  
  pMode(Pin, OUTPUT);    // bring low for 480us

  _delay_us(480);
  
  
  // Release line and wait for 60us
  dWrite(Pin, HIGH);
  pMode(Pin, INPUT);
  _delay_us(60);

  // store line value and wait for completion of 480us period
  i=dRead(Pin);

  _delay_us(480);
  
  // return measured line value
  return i;

}

uint8_t therm_reset()
{
  uint8_t i;

  // pull line low and wait for 480us
  THERM_LOW();
  THERM_OUTPUT_MODE();
  _delay_us(480);
 
  // Release line and wait for 60us
  THERM_INPUT_MODE();
  _delay_us(60);

  // store line value and wait for completion of 480us period
  i=(THERM_PIN & (1<<THERM_DQ));
  _delay_us(480);
    
  // return measured line value
  return i;

}



void therm_write_bit(uint8_t bit)
{
  // pull line low for 1us
  THERM_LOW();
  THERM_OUTPUT_MODE();
  _delay_us(1);
  
  // if we want to write 1, release the line (if not will keep low)
  if (bit) THERM_INPUT_MODE();  

  // wait for 60us and release the line
  _delay_us(60);
  THERM_INPUT_MODE();

}

uint8_t therm_read_bit(void)
{
  uint8_t bit=0;

  // pull line low for 1us
  THERM_LOW();
  THERM_OUTPUT_MODE();
  _delay_us(1);

  // release line and wait for 14us
  THERM_INPUT_MODE();
  _delay_us(14);

  // read line value
  if (THERM_PIN&(1<<THERM_DQ)) bit=1;

  // wait for 45us to end and return read value
  _delay_us(45);

  return bit;

}

uint8_t therm_read_byte(void)
{

  uint8_t i=8, n=0;
 
  while(i--)
  {
    // shift one position right and store read value
    n>>=1;
    n|=(therm_read_bit()<<7);
  }

  return n;
}

void therm_write_byte(uint8_t byte)
{
  uint8_t i=8;

  while(i--)
  {
    // write actual bit and shift one position right to make the next bit ready
    therm_write_bit(byte&1);
    byte>>=1;
  }

}



unsigned char therm_rom_search(unsigned char diff, unsigned char *id)
{
  // scans a one-wire bus for all attached DS18x20 sensors

  unsigned char i, j, next_diff;
  unsigned char b;

  
  if( therm_reset() )
    return PRESENCE_ERR;                       // error, no device found
    
  therm_write_byte( THERM_CMD_SEARCHROM );     // ROM search command
  next_diff = LAST_DEVICE;                     // unchanged on last device
  i = 8 * 8;                        // 8 bytes
  do
  {
    j = 8;                          // 8 bits
    do
    {
      b = therm_read_bit();      // read bit

      // read complement bit
      if( therm_read_bit() ) 
      {         
        if( b )                     // 11
          return DATA_ERR;          // data error
      } else {
        if( !b )
        {
          // 00 = 2 devices
          if( diff > i || ((*id & 1) && diff != i) )
          {
            b = 1;                  // now 1
            next_diff = i;          // next pass 0
          }
        }
      }
      therm_write_bit( b );         // write bit
      *id >>= 1;
      if( b )                       // store bit
        *id |= 0x80;
      i--;
    } while( --j );

    id++;                           // next byte
  } while( i );

  return next_diff;                 // to continue search

} 




void start_meas( void )
{
  if( PINB & 1<< THERM_DQ )
  {
    therm_write_byte(THERM_CMD_CONVERTTEMP);    // read scratchpad

    PORTB |= 1<< THERM_DQ;
    DDRB |= 1<< THERM_DQ;     // parasite power on

  } else {

    // Short circuit on bus
    _delay_ms(1000);
  }
}


int scanbus(void)
{

  // outputs the number of found devices and stores all roms in an array
  unsigned char id[8], diff;
  unsigned char i,x;
  unsigned char device_counter=0;

  // clear array

  for (x=1; x <= ROMS; x++)
  {
    romcodes[x][0] = 0;
  }


  for( diff = SEARCH_FIRST; diff != LAST_DEVICE; )
  {
    diff = therm_rom_search( diff, id );

    if( diff == PRESENCE_ERR )
    {
      //Serial.println("scanbus: no sensors found");
                 
      _delay_ms(1000);
      break;
    }
    if( diff == DATA_ERR )
    {
      //Serial.println("scanbus: Bus error");
      _delay_ms(1000);
      
      break;
    }

    if( id[0] == 0x28 || id[0] == 0x10 )
    {
      // valid temperature sensor found
     
      // store romcode in array
      for( i = 0; i < 8; i++ )
      {

        romcodes[device_counter][i]=id[i];       

      }

      device_counter=device_counter+1;

    }
  }

  return device_counter;
}

void therm_romtemp(uint8_t device)
{
  // address a specific sensor and retrieve its temperature

  uint8_t i;

  therm_reset();
  therm_write_byte(THERM_CMD_MATCHROM);
  for (i=0 ; i < 8; i++)
  {
    therm_write_byte(romcodes[device][i]);
  }
  therm_write_byte(THERM_CMD_CONVERTTEMP);

  // wait until conversion is complete
  while(!therm_read_bit());



  therm_reset();
  therm_write_byte(THERM_CMD_MATCHROM);
  for (i=0 ; i < 8; i++)
  {
    therm_write_byte(romcodes[device][i]);
  }
  therm_write_byte(THERM_CMD_RSCRATCHPAD);

  // store the replies in the 'scratchpad' variable
  for (i=0 ; i < 2; i++)
  {
    scratchpad[i]=therm_read_byte();
  }
  therm_reset();

}


uint16_t showdevs()
{
  // outputs all device romcodes stored in the array

  unsigned char device=0;
  unsigned char rombyte=0;
  uint8_t digit;
  uint16_t decimal;
 
  while (1)
  {

    if (romcodes[device][0] != 0)
    {

      // valid romcode found in array 

      // target the sensor and read its temperature
      therm_romtemp(device);   // this fills the 'scratchpad' variable

      //for (rombyte=0; rombyte < 8; rombyte++)
      //{        
      //  //Serial.print(romcodes[device][rombyte],HEX); // display the romcode 
      //  //Serial.print(" ");
      //} 

       // store temperature integer digits and decimal digits


      // display the 'scratchpad' variable fields

      // --half degree accuracy --   (9 bit)
      // test if temperature is .0 or .5
      //decimal = scratchpad[0] & (1 << 0);  // decimals is in bit 0, store value
      // move temperature bits on position to the right
      //scratchpad[0]>>=1;
      //Serial.print(scratchpad[0]);
      //if (decimal == 0)
      //  Serial.print(".0");
      //else 
      //  Serial.print(".5");



      // -- 2-decimal of a degree accuracy -- (12 bit)
      // store whole digits
      digit=scratchpad[0]>>4;
      digit|=(scratchpad[1]&0x7)<<4;
      // store decimal digits
      decimal=scratchpad[0]&0xf;
      decimal*=THERM_DECIMAL_STEPS_12BIT;

    
      
      
      // decimals are stored as a 4-digit positive number (e.g. 3625), scale back to < 10
      decimal = (int) (decimal / 1000); // eg. from 3625 just keep the '3'

      //multiply digits * 10 and add decimal  (e.g. 253 for 25.3 deg C)
      decimal = (digit * 10) + decimal; 
      
      device=device+1;

      break; // only process one sensor

    } else {

      break;

    }

  } 

  return decimal;
}

//--end DS18B20 functions--



// system wakes up when watchdog is timed out
void system_sleep() 
{

  setup_watchdog(WDTIME);   
  
  noInterrupts();

  ADCSRA &= ~( 1 <<ADEN);    // Disable ADC
  ACSR    =  ( 1 <<ACD);     // Disable the analog comparator

  set_sleep_mode(SLEEP_MODE_PWR_DOWN); // set type of sleep mode

  sleep_enable();  

  interrupts();                         // Enable the Interrupts so the wdt can wake us up

  sleep_mode();                        // System sleeps here

  sleep_disable();                     // System continues execution here when watchdog timed out 
  
  sbi(ADCSRA,ADEN);                    // switch Analog to Digitalconverter ON
}


void setup_watchdog(int ii) 
{

  uint8_t bb;
  if (ii > 9 ) ii=9;
  bb=ii & 7;
  if (ii > 7) bb|= (1<<5);
  bb|= (1<<WDCE);

  MCUSR &= ~(1<<WDRF);
  // start timed sequence
  WDTCR |= (1<<WDCE) | (1<<WDE);
  // set new watchdog timeout value
  WDTCR = bb;
  WDTCR |= _BV(WDIE);
}

void blinkled()
{
  digitalWrite(LEDPIN,HIGH);
  delay(20);
  digitalWrite(LEDPIN,LOW);
  delay(20);
  
}

void loop() 
{
 
 
  // DS18B20 init
  uint8_t  ack; // whether any devices could be detected on onewire bus
  int devs;  // nr of devices discovered on onewire bus
  
     
  uint16_t data_value =  0;  // either 10-bit ADC or Temperature x 10 (e.g. 243 for 24.3 deg C)
  
  uint8_t val1, val2, val3, tmpvar8;
  uint16_t tmpvar16;      

  // prepare 3-Byte (24-bit) code to be sent over the air with the RF module
  char sendstring[24] = { 0 }; // all elements to 0  
  char array1[9] = { 0 };  // 9th is the '/0' closing char of the array
  char array2[9] = { 0 };
  char array3[9] = { 0 };
  
  int pos;
    

  
  data_value = -1; 
  
  if (data_type == 1)
  {
    
    // read from ADC 
    
    digitalWrite(SWITCH,HIGH);  // enable fet(s) to switch on voltage-divider circuit    
    data_value = analogRead(ADCPIN); // value somewhere between 0 - 1023  
    digitalWrite(SWITCH,LOW); // turn fet switch off
    
  } else {
    
    // read from temperature sensor using my own DS18B20 code
    
    ack = owReset(THERM_DQ);

    if (ack == 1)
    {

      // no sensors found   
      
    } else {

      // sensor activity detected       
      
      devs=scanbus();  // detect and store device rom's in an array       
            
      if (devs > 0)
      {

        data_value=showdevs();
       
      } else {
       
        // unable to get sensible data back from sensors
 
      }

    }

    if (data_value==850)
      data_value=-1;
       
  }

    
    
  // break up device_id into 8 + 4 bits
  tmpvar16 = device_id;
  val1 = (int) (tmpvar16 >> 4);  // right-shift by 4 so we'll just have the upper 8 bits.

 

  tmpvar16 = device_id;
  val2 = (int) (tmpvar16) ; // only grab 8 LSB
  val2 = (val2 << 4);  // shift up 4


  // toggle between sending voltage or temperature
  if (data_type == 1)
    data_type = 0;
  else
    data_type = 1;

  
  // shift data_type 2-left and add to val2
  val2 = val2 + (data_type << 2);

  // split up data_value
  tmpvar16 = data_value;
  tmpvar8 = (int) (tmpvar16 >> 8); // right-shift-8 to just get the upper two bits

  // combine with val2
  val2 = val2 + tmpvar8;

  tmpvar16 = data_value;
  val3 = (int) tmpvar16; // only grab 8 LSB

  // val1, val2 and val3 are now set, let's build a single array from these three values

  
  // set array1
  pos = 0;
  while (pos < 8)
  {
    (val1 & (1 << (7-pos))) ? array1[pos]='1' : array1[pos]='0';
    pos++;    
  }
  array1[8]=0;

  // set array2
  pos = 0;
  while (pos < 8)
  {
    (val2 & (1 << (7-pos))) ? array2[pos]='1' : array2[pos]='0';
    pos++;    
  }
  array2[8]=0;

  // set array3
  pos = 0;
  while (pos < 8)
  {
    (val3 & (1 << (7-pos))) ? array3[pos]='1' : array3[pos]='0';
    pos++;    
  }
  array3[8]=0;

  // concatenate the three arrays into one
  memcpy(sendstring,array1,8);
  memcpy(sendstring+8,array2,8);
  memcpy(sendstring+8+8,array3,8);
  
  
  sleepcycle = SLEEPCYCLES; 

 
  blinkled(); // blink before we transmit

  // don't transmit if no data was acquired
  if (data_value != -1)
    mySwitch.send(sendstring);


  // enter sleep cycles (to extend WDT beyond 8 seconds) 
  while (sleepcycle > 1)
  {   
    system_sleep();  // start power down sequence
    sleepcycle --;
  }

  // wakey wakey..
    


  // return to the top
  
}


//This runs each time the watchdog timer wakes us up from sleep
ISR(WDT_vect) {
}
